from dune.perftool.ufl.modified_terminals import Restriction

from dune.perftool.options import get_option

from dune.perftool.pdelab.argument import name_coefficientcontainer
from dune.perftool.pdelab.geometry import world_dimension
from dune.perftool.generation import (class_member,
                                      domain,
                                      function_mangler,
                                      get_global_context_value,
                                      iname,
                                      include_file,
                                      initializer_list,
                                      instruction,
                                      loopy_class_member,
                                      preamble,
                                      silenced_warning,
                                      temporary_variable,
                                      transform,
                                      valuearg
                                      )
from dune.perftool.loopy.buffer import get_buffer_temporary
from dune.perftool.loopy.vcl import ExplicitVCLCast
from dune.perftool.pdelab.localoperator import (name_domain_field,
                                                lop_template_range_field,
                                                )
from dune.perftool.pdelab.quadrature import quadrature_order
from dune.perftool.tools import maybe_wrap_subscript
from loopy import CallMangleInfo
from loopy.symbolic import FunctionIdentifier
from loopy.types import NumpyType

from pytools import ImmutableRecord, product

import pymbolic.primitives as prim
import loopy as lp
import numpy as np


def ceildiv(a, b):
    return -(-a // b)


class BasisTabulationMatrixBase(object):
    pass


class BasisTabulationMatrix(BasisTabulationMatrixBase, ImmutableRecord):
    def __init__(self,
                 quadrature_size=None,
                 basis_size=None,
                 transpose=False,
                 derivative=False,
                 face=None,
                 slice_size=None,
                 slice_index=None,
                 ):
        if quadrature_size is None:
            quadrature_size = quadrature_points_per_direction()
        if slice_size is not None:
            quadrature_size = ceildiv(quadrature_size, slice_size)
        if basis_size is None:
            basis_size = basis_functions_per_direction()
        ImmutableRecord.__init__(self,
                                 quadrature_size=quadrature_size,
                                 basis_size=basis_size,
                                 transpose=transpose,
                                 derivative=derivative,
                                 face=face,
                                 slice_size=slice_size,
                                 slice_index=slice_index,
                                 )

    @property
    def rows(self):
        if self.transpose:
            return self.basis_size
        else:
            return self.quadrature_size

    @property
    def cols(self):
        if self.transpose:
            return self.quadrature_size
        else:
            return self.basis_size

    def pymbolic(self, indices):
        name = "{}{}Theta{}{}_qp{}_dof{}".format("face{}_".format(self.face) if self.face is not None else "",
                                                 "d" if self.derivative else "",
                                                 "T" if self.transpose else "",
                                                 "_slice{}".format(self.slice_index) if self.slice_size is not None else "",
                                                 self.quadrature_size,
                                                 self.basis_size,
                                                 )
        define_theta(name, self)
        return prim.Subscript(prim.Variable(name), indices)

    @property
    def vectorized(self):
        return False


class BasisTabulationMatrixArray(BasisTabulationMatrixBase):
    def __init__(self, tabs, width=None):
        assert isinstance(tabs, tuple)

        # Assert that all the basis tabulations match in size!
        assert len(set(t.quadrature_size for t in tabs)) == 1
        assert len(set(t.basis_size for t in tabs)) == 1
        assert len(set(t.transpose for t in tabs)) == 1
        assert len(set(t.face for t in tabs)) == 1
        assert len(set(t.slice_size for t in tabs)) == 1
        self.tabs = tabs

        if width is None:
            width = len(tabs)
        self.width = width

    @property
    def quadrature_size(self):
        return self.tabs[0].quadrature_size

    @property
    def basis_size(self):
        return self.tabs[0].basis_size

    @property
    def transpose(self):
        return self.tabs[0].transpose

    @property
    def face(self):
        return self.tabs[0].face

    @property
    def derivative(self):
        return tuple(t.derivative for t in self.tabs)

    @property
    def slice_size(self):
        return self.tabs[0].slice_size

    @property
    def slice_index(self):
        return tuple(t.slice_index for t in self.tabs)

    @property
    def rows(self):
        if self.transpose:
            return self.basis_size
        else:
            return self.quadrature_size

    @property
    def cols(self):
        if self.transpose:
            return self.quadrature_size
        else:
            return self.basis_size

    def pymbolic(self, indices):
        assert len(indices) == 3

        # Check whether we can realize this by broadcasting the values of a simple tabulation
        if len(set(self.tabs)) == 1:
            theta = self.tabs[0].pymbolic(indices[:-1])
            return prim.Call(ExplicitVCLCast(np.float64, vector_width=len(self.tabs)), (theta,))

        abbrevs = tuple("{}x{}".format("d" if t.derivative else "",
                                       "s{}".format(t.slice_index) if t.slice_size is not None else "")
                        for t in self.tabs)
        name = "ThetaLarge{}{}_{}_qp{}_dof{}".format("face{}_".format(self.face) if self.face is not None else "",
                                                     "T" if self.transpose else "",
                                                     "_".join(abbrevs),
                                                     quadrature_points_per_direction(),
                                                     basis_functions_per_direction(),
                                                     )
        for i, tab in enumerate(self.tabs):
            define_theta(name, tab, additional_indices=(i,), width=self.width)

        # Apply padding to those fields not used. This is necessary because you may get memory
        # initialized with NaN and those NaNs will screw the horizontal_add.
        i = theta_iname("i", self.rows)
        j = theta_iname("j", self.cols)
        for index in set(range(self.width)) - set(range(len(self.derivative))):
            instruction(assignee=prim.Subscript(prim.Variable(name), (prim.Variable(i), prim.Variable(j), index)),
                        expression=0,
                        kernel="operator")

        member = loopy_class_member(name,
                                    classtag="operator",
                                    dtype=np.float64,
                                    dim_tags="f,f,vec",
                                    shape=(self.rows, self.cols, self.width),
                                    potentially_vectorized=True,
                                    managed=True,
                                    )

        return prim.Subscript(prim.Variable(member), indices)

    @property
    def vectorized(self):
        return True


def quadrature_points_per_direction():
    # Quadrature order
    q = quadrature_order()

    # Quadrature points in per direction
    nb_qp = q // 2 + 1

    return nb_qp


def polynomial_degree():
    form = get_global_context_value("formdata").preprocessed_form
    return form.coefficients()[0].ufl_element()._degree


def basis_functions_per_direction():
    return polynomial_degree() + 1


def define_oned_quadrature_weights(name):
    loopy_class_member(name,
                       dtype=np.float64,
                       classtag="operator",
                       shape=(quadrature_points_per_direction(),),
                       )


def name_oned_quadrature_weights():
    num = quadrature_points_per_direction()
    name = "qw_num{}".format(num)
    define_oned_quadrature_weights(name)
    return name


def define_oned_quadrature_points(name):
    loopy_class_member(name,
                       dtype=np.float64,
                       classtag="operator",
                       shape=(quadrature_points_per_direction(),),
                       )


def name_oned_quadrature_points():
    num = quadrature_points_per_direction()
    name = "qp_num{}".format(num)
    define_oned_quadrature_points(name)
    return name


@class_member(classtag="operator")
def typedef_polynomials(name, degree):
    range_field = lop_template_range_field()
    domain_field = name_domain_field()

    include_file("dune/pdelab/finiteelementmap/qkdg.hh", filetag="operatorfile")

    # TODO: make switching polynomials possible
    # return "using {} = Dune::QkStuff::GaussLobattoLagrangePolynomials<{}, {}, {}>;".format(name,
    #                                                                                        domain_field,
    #                                                                                        range_field,
    #                                                                                        degree)
    return "using {} = Dune::QkStuff::EquidistantLagrangePolynomials<{}, {}, {}>;".format(name,
                                                                                          domain_field,
                                                                                          range_field,
                                                                                          degree)


def type_polynomials(degree):
    name = "Polynomials1D_Degree{}".format(degree)
    typedef_polynomials(name, degree)
    return name


@class_member(classtag="operator")
def define_polynomials(name, degree):
    polynomials_type = type_polynomials(degree)
    return "{} {};".format(polynomials_type, name)


def name_polynomials(degree):
    name = "poly_degree{}".format(degree)
    define_polynomials(name, degree)
    return name


@preamble(kernel="operator")
def sort_quadrature_points_weights(qp, qw):
    range_field = lop_template_range_field()
    domain_field = name_domain_field()
    number_qp = quadrature_points_per_direction()
    include_file("dune/perftool/sumfact/onedquadrature.hh", filetag="operatorfile")
    return "onedQuadraturePointsWeights<{}, {}, {}>({}, {});".format(range_field, domain_field, number_qp, qp, qw)


@iname(kernel="operator")
def theta_iname(name, bound):
    name = "{}_{}".format(name, bound)
    domain(name, bound, kernel="operator")
    return name


class PolynomialLookup(FunctionIdentifier):
    def __init__(self, pol, derivative):
        self.pol = pol
        self.derivative = derivative

    def __getinitargs__(self):
        return (self.pol, self.derivative)

    @property
    def name(self):
        return "{}.{}".format(self.pol, "dp" if self.derivative else "p")


@function_mangler
def polynomial_lookup_mangler(target, func, dtypes):
    if isinstance(func, PolynomialLookup):
        return CallMangleInfo(func.name, (NumpyType(np.float64),), (NumpyType(np.int32), NumpyType(np.float64)))


def define_theta(name, tabmat, additional_indices=(), width=None):
    assert isinstance(tabmat, BasisTabulationMatrix)
    qp = name_oned_quadrature_points()
    qw = name_oned_quadrature_weights()
    sort_quadrature_points_weights(qp, qw)
    degree = tabmat.basis_size - 1
    polynomials = name_polynomials(degree)

    shape = (tabmat.rows, tabmat.cols)
    dim_tags = "f,f"
    if additional_indices:
        dim_tags = dim_tags + ",c"
        shape = shape + (width,)

    loopy_class_member(name,
                       dtype=np.float64,
                       shape=shape,
                       classtag="operator",
                       dim_tags=dim_tags,
                       managed=True,
                       potentially_vectorized=True,
                       )

    i = prim.Variable(theta_iname("i", tabmat.rows))
    j = prim.Variable(theta_iname("j", tabmat.cols))

    inames = [i, j]
    if tabmat.transpose:
        inames = [j, i]

    if tabmat.slice_size is not None:
        inames[0] = tabmat.slice_size * inames[0] + tabmat.slice_index

    args = [inames[1], prim.Subscript(prim.Variable(qp), (inames[0],))]
    if tabmat.face is not None:
        args[1] = tabmat.face

    instruction(assignee=prim.Subscript(prim.Variable(name), (i, j) + additional_indices),
                expression=prim.Call(PolynomialLookup(polynomials, tabmat.derivative), tuple(args)),
                kernel="operator",
                )


def construct_basis_matrix_sequence(transpose=False, derivative=None, facedir=None, facemod=None, basis_size=None):
    dim = world_dimension()
    result = [None] * dim

    for i in range(dim):
        quadrature_size = quadrature_points_per_direction()
        if basis_size is None:
            basis_size = basis_functions_per_direction()

        onface = None
        if facedir == i:
            quadrature_size = 1
            onface = facemod

        result[i] = BasisTabulationMatrix(quadrature_size=quadrature_size,
                                          basis_size=basis_size,
                                          transpose=transpose,
                                          derivative=derivative == i,
                                          face=onface)

    return tuple(result)
